#include "fixed.hpp"
#include <assert.h>

unsigned long longMath::uldiv (unsigned long a, unsigned long b)
{
unsigned long abits,bbits;

assert (b != 0);

abits = (a/b);
abits = abits << 16;

// a%b < 1
if ((a>>16) == 0 && (b>>16) == 0)
	{
	bbits = a%b;
	bbits = bbits <<16;
	bbits = bbits / b;
	}
else if (b <= 0xFFFFL)
	{
	if (b > CLEARHIBITS(a) || b > 6553L)
		{
		bbits = a%b;
		bbits = bbits << 16;
		bbits = bbits /b;
		}
	else
		{
		// error here!
		// Error is largest when the denominator is small (>.001)
		// 0 < error < 100+
		bbits = 0;
		unsigned long digit;
		int x;
		int dividor = 10;

		for (x = 0; x < 5; x++)
			{
			a *= 10 * ((65535L/(b*10)));
			a = CLEARHIBITS(a);
			digit = a/b;
			bbits += (digit*65535L)/(dividor);
			dividor *= 10;
			b *= 10;
//			b = b << 2;
			if (bbits > 0xFFFFL)
				bbits = bbits >> 1;
			}
		}
	}
// a < b
// b has a fractional part

/*
TO:  Generate digits
	  1.  Calc the modulus
	  2.  mod *= 10
	  3.  digit = (mod / b) /10
	  4.  mod = mod % b
*/
// a > b
// a % b > 1
// a or b have a fractional part
else
	{
	bbits = 0;
	unsigned long mod, digit;
	int x;
	int dividor = 10;

	mod = a%b;  // mod is > 1; fraction = mod/b
	for (x = 0; x < 5; x++)
		{
		mod *= 10;
		digit = (mod/b);
		bbits += (digit*65535L)/(dividor);
		dividor *= 10;
		mod = mod % b;
		}
	}
return abits + bbits;
}

bool longMath::operator == (longMath &f)
{return (lNum == f.lNum) ? true : false;}

bool longMath::operator <= (int i)
{return ((lNum) <= (((long)i)<<16)) ? true : false;}

bool longMath::operator > (int i)
{return ((lNum) > (((long)i)<<16)) ? true : false;}

ufixed ufixed::fResult;
fixed fixed::fResult;

ufixed &ufixed::sqrt()
{
unsigned long s1,s2;

s1 = lNum/2;
for(;;)
	{
	s2 = lNum/s1;
	s2 = s2+s1;
	s2 = s2/2;
	if (s1 <= s2)
		{
		break;
		}
	s1 = s2;
	}
fResult.lNum = s2;
return fResult;
}

ufixed &ufixed::operator + (double f)
{fResult.lNum = ((unsigned long)lNum + (unsigned long)( f * _BITS_ ));return fResult;}

ufixed &ufixed::operator - (double f)
{fResult.lNum  = ((unsigned long)lNum - (unsigned long)( f * _BITS_ ));return fResult;}

ufixed &ufixed::operator * (ufixed &f)
{fResult.lNum  = (ulmult(lNum, f.lNum));return fResult;}

ufixed &ufixed::operator * (double f)
{fResult.lNum  = (ulmult(lNum, (unsigned long)( f * _BITS_ )));return fResult;}

ufixed &ufixed::operator / (ufixed &f)
{fResult.lNum  = (uldiv(lNum, f.lNum));return fResult;}

ufixed &ufixed::operator / (double f)
{fResult.lNum  = (uldiv(lNum, (unsigned long)( f * _BITS_ )));return fResult;}

ufixed &ufixed::operator = (ufixed &f)
{lNum = (unsigned long)f.lNum; return *this;}

ufixed &ufixed::operator = (unsigned long f)
{lNum = (unsigned long)((double)f)*_BITS_; return *this;}

ufixed &ufixed::operator = (double f)
{lNum = (unsigned long)( f * _BITS_ );return *this;}

ufixed::operator double ()
{
return (double) ( ((double) ( (unsigned long) lNum) ) / _BITS_ );
}

ufixed::operator int ()
{return (int)(((double)lNum) / _BITS_);}

/*
	fixed
sqrt:
s1&=x&/2
L01:
s2&=x&/s1&
s2&=s2&+s1&
s2&=s2&/2
if s1&>s2& then
	s1&=s2&
	goto L01:
endif
x&=s2&
return
*/

fixed &fixed::sqrt()
{
unsigned long s1,s2;

if (lNum > 0)		// no imaginary numbers and no error handling!
	{
	fResult.lNum = -1;
	return fResult;
	}
s1 = lNum/2;
for(;;)
	{
	s2 = lNum/s1;
	s2 = s2+s1;
	s2 = s2/2;
	if (s1 <= s2)
		{
		break;
		}
	s1 = s2;
	}
fResult.lNum = s2;
return fResult;
}

fixed &fixed::operator + (double f)
{fResult.lNum  = (lNum + ( f * _BITS_ ));return fResult;}

fixed &fixed::operator - (fixed &f)
{fResult.lNum  = (lNum - f.lNum);return fResult;}

fixed &fixed::operator - (double f)
{fResult.lNum  = (lNum - (long)( f * _BITS_ ));return fResult;}


fixed &fixed::operator * (double f)
{
uchar s1,s2;
s1 = lNum>>31;
s2 = (f < 0) ? 1 : 0;
if (s1 != 0)
	lNum = -lNum;
if (s2 != 0)
	f *= (-1);
fResult.lNum  = (ulmult(lNum, (long)( f * _BITS_ )));
if ((s1 == 0 && s2 == 0)||(s1 == 1 && s2 == 1))
	return fResult;
fResult.lNum = -fResult.lNum;
return fResult;
}

fixed &fixed::operator / (fixed &f)
{
uchar s1,s2;
s1 = lNum>>31;
s2 = f.lNum>>31;
if (s1 != 0)
	lNum = -lNum;
if (s2 != 0)
	f.lNum = -f.lNum;
fResult.lNum  = (uldiv(lNum, f.lNum));
if (s2 != 0)
	f.lNum = -f.lNum;
if ((s1 == 0 && s2 == 0)||(s1 == 1 && s2 == 1))
	return fResult;
fResult.lNum = -fResult.lNum;
return fResult;
}

fixed &fixed::operator / (double f)
{
uchar s1,s2;
s1 = lNum>>31;
s2 = (f < 0) ? 1 : 0;
if (s1 != 0)
	lNum = -lNum;
if (s2 != 0)
	f = -f;
fResult.lNum  = (uldiv(lNum, (long)( f * _BITS_ )));
if ((s1 == 0 && s2 == 0)||(s1 == 1 && s2 == 1))
	return fResult;
fResult.lNum = -fResult.lNum;
return fResult;
}

fixed::operator double ()
{
return (double)(((double)lNum) / _BITS_);
}

fixed::operator int ()
{
return (int)(((long)lNum) / _BITS_);
}

