#ifndef _fixed_
#define _fixed_

#include "common.h"

#define _BITSHIFT_ 16
#define _BITS_  (1L << _BITSHIFT_)
#define _LONG2DBL_ ( a )	((double)a / _BITS_)
#define _DBL2LONG_ ( a )	((long)( a * _BITS_ ))

#define CLEARHIBITS(a) (a & 0xFFFFL)
#define CLEARLOWBITS(a) (a & 0xFFFF0000L)

//#define _ADD(a, b)	_LONG2DBL_ (( _DBL2LONG_(a) + _DBL2LONG_(b) ))
//#define _SUB(a, b)	_LONG2DBL_ (( _DBL2LONG_(a) - _DBL2LONG_(b) ))
//#define _MUL(a, b)	_LONG2DBL_ (( _DBL2LONG_(a) * _DBL2LONG_(b) ) >> _BITSHIFT_))
//#define _DIV(a, b)	_LONG2DBL_ (( _DBL2LONG_(a) / _DBL2LONG_(b) ) << _BITSHIFT_))

class longMath : public generic {
	protected:
		long lNum;
		unsigned long ulmult(unsigned long, unsigned long);
		unsigned long uldiv (unsigned long, unsigned long);
	public:
		longMath operator + (longMath &f) const;
		longMath operator + (double f) const;
		longMath operator - (longMath &f) const;
		longMath operator - (double f) const;
		longMath operator * (longMath &f) const;
		longMath operator * (double f) const;
		longMath operator / (longMath &f) const;
		longMath operator / (double f) const;

		bool operator == (longMath &f);
		bool operator <= (longMath &);
		bool operator <= (int);
		bool operator > (int);
		bool operator > (longMath &);
};

class ufixed : public longMath {
	protected:
		static ufixed fResult;
	public:
		ufixed &sqrt ();

		ufixed &operator + (ufixed &f);
		ufixed &operator + (double f);
		ufixed &operator - (ufixed &f);
		ufixed &operator - (double f);
		ufixed &operator * (ufixed &f);
		ufixed &operator * (double f);
		ufixed &operator / (ufixed &f);
		ufixed &operator / (double f);
		ufixed &operator = (ufixed &f);
		ufixed &operator = (unsigned long f);
		ufixed &operator = (double);

		operator double ();
		operator int ();
};

class fixed : public longMath {
	protected:
		static fixed fResult;
	public:
		fixed &sqrt ();

		fixed &operator + (fixed &f);
		fixed &operator + (double f);
		fixed &operator - (fixed &f);
		fixed &operator - (double f);
		fixed &operator * (fixed &f);
		fixed &operator * (double f);
		fixed &operator / (fixed &f);
		fixed &operator / (double f);
		fixed &operator = (fixed &f);
		fixed &operator = (long f);
		fixed &operator = (double);

		operator double ();
		operator int ();
};

inline unsigned long longMath::ulmult (unsigned long a, unsigned long b)
{
unsigned long ia, ib, fa, fb, ir;
ia = a>>16;
ib = b>>16;
fa = CLEARHIBITS(a);
fb = CLEARHIBITS(b);


ir = (ia*ib)<<16;//*65535L;
return ir + (ia*fb)+(fa*ib)+(((fa*fb)>>16)&0XFFFFL);
}

inline fixed &fixed::operator + (fixed &f)
{
fResult.lNum  = (lNum + f.lNum);
return fResult;
}

inline fixed &fixed::operator * (fixed &f)
{
uchar s1, s2;
s1 = lNum>>31;
s2 = f.lNum>>31;
if (s1 != 0)
	lNum = -lNum;
if (s2 != 0)
	f.lNum = -f.lNum;
fResult.lNum = (ulmult(lNum, f.lNum));
if (s2 != 0)
	f.lNum = -f.lNum;
if ((s1 == 0 && s2 == 0)||(s1 == 1 && s2 == 1))
	return fResult;
fResult.lNum = -fResult.lNum;
return fResult;
}

inline ufixed &ufixed::operator + (ufixed &f)
{fResult.lNum = ((unsigned long)lNum + (unsigned long)f.lNum);return fResult;}

inline ufixed &ufixed::operator - (ufixed &f)
{fResult.lNum  = ((unsigned long)lNum - (unsigned long)f.lNum);return fResult;}

inline fixed &fixed::operator = (fixed &f)
{
lNum = (long)f.lNum;
return *this;
}

inline fixed &fixed::operator = (long f)
{
lNum = (long)((double)f)*_BITS_;
return *this;
}

inline fixed &fixed::operator = (double f)
{
lNum = (long)( f * _BITS_ );
return *this;
}

#endif
